<#
.SYNOPSIS
    Migrates blobs from SQL databases to Azure Blob Storage.

.DESCRIPTION
    Migrates blobs from SQL databases to Azure Blob Storage.

.PARAMETER SqlDBConnectionNames
    An array which contains names of SQL databases connection strings which should be migrated. 
    The connection string names should exist in the connection strings config file. You can specify the file using -ConnectionStringsConfigFileRelativePath parameter. 
    If you don't specify -ConnectionStringsConfigFileRelativePath, the default value which is "..\App_Config\ConnectionStrings.Config" will be used to find the connection strings.

.PARAMETER AzureBlobConnectionName
    Name of the connection string to the Azure blob storage. 
    The connection string name should exist in the connection strings config file. You can specify the file using -ConnectionStringsConfigFileRelativePath parameter. 
    If you don't specify -ConnectionStringsConfigFileRelativePath, the default value which is "..\App_Config\ConnectionStrings.Config" will be used to find the connection strings.

.PARAMETER BlobContainerName
    Name of the container on the storage account which contains the blobs. The container should exists on the storage. The default value is "blobcontainer".

.PARAMETER ConnectionStringsConfigFileRelativePath
    Relative path to the connection strings config file. The default value is "..\..\App_Config\ConnectionStrings.Config"

.EXAMPLE
    .\Invoke-Migration.ps1 -SqlDBConnectionNames @("master", "web") -AzureBlobConnectionName "azure" -BlobContainerName "blobs" -Force
#>
[CmdletBinding()]
param(
    [Parameter(Mandatory=$True)]
    [ValidateNotNullOrEmpty()]
    [string[]]
    $SqlDBConnectionNames,

    [Parameter(Mandatory=$True)]
    [ValidateNotNullOrEmpty()]
    [string]
    $AzureBlobConnectionName,

    [Parameter(Mandatory=$True)]
    [ValidateNotNullOrEmpty()]
    [string]
    $BlobContainerName,
    
    [ValidateNotNullOrEmpty()]
    [string]
    $ConnectionStringsConfigFileRelativePath = "..\..\App_Config\ConnectionStrings.Config",

    [switch] $Force
)

#region Common Helper Functions

<#
Writes the information to the running console. 
#>
Function Write-Info([object] $Info)
{
    Write-Output $Info
}

<#
    Retry Strategy for the Retry-Block function
#>
Enum RetryStrategy
{
    #Delay *= 2
    Exponential

    #Delay = InitialDelay
    Fixed

    #Delay += InitialDelay
    Incremental
}

<#
    This function disposes the input object after executing the code block regardless of having or not having exceptions. Works like using block in C#. 
#>
Function Using-Object(
    [System.IDisposable]
    $InputObject,
    
    [scriptblock]
    $ScriptBlock = {throw "ScriptBlock is mandatory, please provide a value."})
{
    try
    {
        . $ScriptBlock
    }
    finally
    {
        if ($null -ne $InputObject -and $InputObject -is [System.IDisposable])
        {
            $InputObject.Dispose()
        }
    }
}

<#
    This function checks if an exception is a transient exception.
#>
Function IsTransient(
    [System.Exception] $Ex)
{
    if ($Ex -eq $null)
    {
        return $False
    }
    if($Ex -is [System.Data.SqlClient.SqlException])
    {
        [System.Data.SqlClient.SqlException] $SqlException = ($Ex -as [System.Data.SqlClient.SqlException])
        $SqlTransientErrorCodes = @(942,4060,4221,40501,10928,10928,10929,10053,10054,10060,40197,40540,40613,49919,49920,40143,49918,233,64,20)
        $Contains =  $SqlException.Errors | where {$SqlTransientErrorCodes -contains $_.Number}
        if($Contains)
        {
            return $true
        }
        else
        {
            [System.ComponentModel.Win32Exception] $Win32Ex = ($SqlException.InnerException -as [System.ComponentModel.Win32Exception])
            if($Win32Ex)
            {
                if(@(0x102, 0x121) -contains $Win32Ex.NativeErrorCode)
                {
                    return $true
                }
            }
        }
    }
    if($Ex -is [System.TimeoutException])
    {
        return $true
    }   
    if($Ex.InnerException -is [System.Data.SqlClient.SqlException])  
    {
        return IsTransient -Ex $Ex.InnerException
    }
    if($Ex.Message -contains "Please connect to internet before executing this cmdlet")
    {
        return $true
    }
    return $false
}

<#
    This function retries executing of the given code block, by the given retry count, delay and strategy.
	The $Delay parameter unit is Seconds.
#>
Function Retry-Block (
    [int]
    $RetryCount = 5,

    [int]
    $Delay = 1,

    [RetryStrategy] $RetryStrategy = ([RetryStrategy]::Exponential),

    [Parameter(Position = 0)]
    [scriptblock]
    $ScriptBlock = {throw "ScriptBlock is mandatory, please provide a value."})
{
    $Attempt = 1
    $Flag = $true
    $CurrentWaitTime = $Delay
    do 
    {
        try 
        {
            Invoke-Command -ScriptBlock $ScriptBlock -OutVariable Result -ErrorAction Stop              
            $Flag = $false
        }
        catch 
        {
            if(!(IsTransient -Ex $_.Exception))
            {
                $Flag = $false
                throw 
            }
            if ($Attempt -gt $RetryCount) 
            {
                $Flag = $false
                throw 
            }
            else 
            {
                switch($RetryStrategy)
                {
                    ([RetryStrategy]::Exponential)
                    {
                        $CurrentWaitTime *= 2
                    }
                    ([RetryStrategy]::Incremental)
                    {
                        $CurrentWaitTime += $Delay
                    }
                }
                Start-Sleep -Seconds $CurrentWaitTime
                $Attempt++
            }
        }
    }
    While ($Flag)
}

#endregion

#region Invoke-Migration Helper Functions

<#
Gets all blob ids page by page
#>
Function GetBlobIdsByPage([string]$ConnectionString, [int] $PageNumber, [int] $PageSize)
{
    $Query = "SELECT DISTINCT [BlobId] FROM [Blobs] ORDER BY [BlobId] ASC OFFSET $(($PageNumber-1)*$PageSize) ROWS FETCH NEXT $PageSize ROWS ONLY"
    Retry-Block {
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                $Connection.Open()
                Using-Object($Reader = $Command.ExecuteReader()) {
                    $Datatable = New-Object System.Data.DataTable
                    $Datatable.Load($Reader)
                    $Datatable
                }        
            }
        }
    }
}

<#
Gets the blob Size in bytes from sql database
#>
Function GetBlobSizeInSql([string]$ConnectionString, [string] $BlobId)
{
    $Query = "SELECT ISNULL(SUM(DATALENGTH([Data])),0) FROM [Blobs] WHERE [BlobId] = '$BlobId'"
    Retry-Block {
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                $Connection.Open()
                $Command.ExecuteScalar()
            }
        }
    }
}

<#
Checks if the blob exists in SQL database
#>
Function BlobExistsInSql([string]$ConnectionString, [string] $BlobId)
{
    $Query = "SELECT COUNT(*) FROM [Blobs] WHERE [BlobId] = '$BlobId'"
    Retry-Block {
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                $Connection.Open()
                $Command.ExecuteScalar()
            }
        }
    }
}

<#
Gets the count of distinct blob ids.
#>
Function GetBlobIdsCount([string]$ConnectionString)
{
    $Query = "SELECT COUNT(DISTINCT [BlobId]) FROM [Blobs]"
    Retry-Block {
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                $Connection.Open()
                $Command.ExecuteScalar()
            }
        }
    }
}

<#
Downloads the blob by id
#>
Function DownloadBlobFromSql([string]$ConnectionString, [string] $BlobId, [string] $Folder)
{
    Retry-Block {
        $Query = "SELECT [Data] FROM [Blobs] WHERE [BlobId] = '$BlobId' ORDER BY [Index]"
        New-Item -ItemType Directory -Force -Path $Folder | Out-Null
        $Extension = "blob"
        $FileName = Join-Path $Folder "$BlobId.$Extension"
        If(Test-Path $FileName)
        {
            Remove-Item $FileName -Force
        } 
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            $Connection.Open()
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                Using-Object($Reader = $Command.ExecuteReader([System.Data.CommandBehavior]::SequentialAccess)) {
                    Using-Object($FileSteam = New-Object System.IO.FileStream($FileName, [System.IO.FileMode]::Create, [System.IO.FileAccess]::Write)) {
                        While($Reader.Read())
                        {
                            Using-Object ($Data = $Reader.GetStream(0)) {
                                $Data.CopyTo($FileSteam)
                            }
                        }
                    }
                }
            }
        }
        $FileName
    }
}

<#
Updates blob ids to URI format
#>
Function UpdateBlobIds([string]$ConnectionString)
{
    $Query = @"
DECLARE @AffectedRows INT = 1;
DECLARE @Total INT = 0;
DECLARE @BatchSize INT = 10;
DECLARE @BlobFields TABLE ([FieldId] UNIQUEIDENTIFIER PRIMARY KEY ([FieldId]));
SET NOCOUNT ON;
INSERT INTO @BlobFields ([FieldId]) (
	SELECT [ItemId] FROM [SharedFields] WHERE [FieldId] = 'FF8A2D01-8A77-4F1B-A966-65806993CD31' AND [Value] = '1' 
		UNION
	SELECT [ItemId] FROM [VersionedFields] WHERE [FieldId] = 'FF8A2D01-8A77-4F1B-A966-65806993CD31' AND [Value] = '1'
		UNION
	SELECT [ItemId] FROM [UnversionedFields] WHERE [FieldId] = 'FF8A2D01-8A77-4F1B-A966-65806993CD31' AND [Value] = '1'
		UNION
	SELECT [ArchivalId] FROM [ArchivedFields] WHERE [FieldId] = 'FF8A2D01-8A77-4F1B-A966-65806993CD31' AND [Value] = '1'
);

BEGIN TRY
	BEGIN TRANSACTION UpdateBlobFields
		WHILE(@AffectedRows>0)
		BEGIN
				UPDATE TOP(@BatchSize) [SharedFields]
				SET [Value] = 'blob://' + REPLACE(REPLACE([Value], '{', ''), '}', '')
				WHERE 
					([FieldId] IN (SELECT [FieldId] FROM @BlobFields)) AND
					([Value] IS NOT NULL) AND 
					([Value] <> '') AND
					(LEN([Value]) <= 38) AND
					(TRY_CONVERT(UNIQUEIDENTIFIER, [Value]) IS NOT NULL)

				SET @AffectedRows = @@ROWCOUNT
				SET @Total = @Total + @AffectedRows
		END

		SET @AffectedRows = 1;
		WHILE(@AffectedRows>0)
		BEGIN
				UPDATE TOP(@BatchSize) [VersionedFields]
				SET [Value] = 'blob://' + REPLACE(REPLACE([Value], '{', ''), '}', '')
				WHERE 
					([FieldId] IN (SELECT [FieldId] FROM @BlobFields)) AND
					([Value] IS NOT NULL) AND 
					([Value] <> '') AND
					(LEN([Value]) <= 38) AND
					(TRY_CONVERT(UNIQUEIDENTIFIER, [Value]) IS NOT NULL)

				SET @AffectedRows = @@ROWCOUNT
				SET @Total = @Total + @AffectedRows
		END

		SET @AffectedRows = 1;
		WHILE(@AffectedRows>0)
		BEGIN
				UPDATE TOP(@BatchSize) [UnversionedFields]
				SET [Value] = 'blob://' + REPLACE(REPLACE([Value], '{', ''), '}', '')
				WHERE 
					([FieldId] IN (SELECT [FieldId] FROM @BlobFields)) AND
					([Value] IS NOT NULL) AND 
					([Value] <> '') AND
					(LEN([Value]) <= 38) AND
					(TRY_CONVERT(UNIQUEIDENTIFIER, [Value]) IS NOT NULL)

				SET @AffectedRows = @@ROWCOUNT
				SET @Total = @Total + @AffectedRows
		END

		SET @AffectedRows = 1;
		WHILE(@AffectedRows>0)
		BEGIN
				UPDATE TOP(@BatchSize) [ArchivedFields]
				SET [Value] = 'blob://' + REPLACE(REPLACE([Value], '{', ''), '}', '')
				WHERE 
					([FieldId] IN (SELECT [FieldId] FROM @BlobFields)) AND
					([Value] IS NOT NULL) AND 
					([Value] <> '') AND
					(LEN([Value]) <= 38) AND
					(TRY_CONVERT(UNIQUEIDENTIFIER, [Value]) IS NOT NULL)

				SET @AffectedRows = @@ROWCOUNT
				SET @Total = @Total + @AffectedRows
		END
	COMMIT TRANSACTION UpdateBlobFields
	SELECT @Total As AffectedRows
END TRY
BEGIN CATCH
	IF @@TRANCOUNT > 0
	BEGIN 
		ROLLBACK TRANSACTION UpdateBlobFields; 
	END;
	THROW
END CATCH
"@
    Retry-Block {
        Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($ConnectionString)) {
            Using-Object ($Command = New-Object System.Data.SQLClient.SQLCommand($Query, $Connection)) {
                $Connection.Open()
                $Command.ExecuteScalar()
            }        
        }
    }
}

<#
Gets the azure container.
#>
Function GetContainer($Context, $ContainerName)
{
    Retry-Block {
        Get-AzureStorageContainer -Name $ContainerName -Context $Context
    }
}

<#
Uploads blob to azure.
#>
Function UploadBlobToAzure($Context, $ContainerName, $File, $BlobId)
{
    Retry-Block {
        Set-AzureStorageBlobContent -Context $Context -Container $ContainerName -File $File -Blob $BlobId -Force 
    }
}

<#
Checks if blob exists in azure.
#>
Function BlobExistsInAzure($Context, $ContainerName, $BlobId)
{
    Retry-Block {
        Get-AzureStorageBlob -Blob $BlobId -Container $ContainerName -Context $Context -ErrorAction Ignore
    }
}
#endregion


$ErrorActionPreference = "Stop"
$ProgressPreference="SilentlyContinue"
Import-Module AzureRM.Storage

$LockFileName = "lock"
$ScriptInvocationPath = Split-Path $MyInvocation.MyCommand.Path
$LockFilePath = (Join-Path $ScriptInvocationPath $LockFileName)
$Lock = $Null
try
{
$Lock = [System.IO.File]::Open($LockFilePath, [System.IO.FileMode]::OpenOrCreate, [System.IO.FileAccess]::Write, [System.IO.FileShare]::None)
$Lock.Write(@(0), 0, 1)
$Lock.Flush()

#region Validations
$SitecoreLogo = @"
  _____ _ _                                                       
 / ____(_) |                                                      
| (___  _| |_ ___  ___ ___  _ __ ___                              
 \___ \| | __/ _ \/ __/ _ \| '__/ _ \                             
 ____) | | ||  __/ (_| (_) | | |  __/                             
|_____/|_|\__\___|\___\___/|_|  \___|                             
                                   
"@
Write-Info $SitecoreLogo
Write-Info "-------------------------------------"
Write-Info "Invoke-Migration started ...         "
Write-Info "$(Get-Date)"
Write-Info "-------------------------------------"
Write-Info "`n"

#Valiates if the connection string file exists
Write-Info "Validating connection string file ..."
$ScriptInvocationPath = Split-Path $MyInvocation.MyCommand.Path
$ConnectionStringsConfigFilePath = (Join-Path $ScriptInvocationPath $ConnectionStringsConfigFileRelativePath)
If(!(Test-Path -Path $ConnectionStringsConfigFilePath -PathType Leaf)) 
{
    Write-Error "'$ConnectionStringsConfigFilePath' is not a valid file path."
}

#Get Connection strings from $ConnectionStringsConfigFilePath
Write-Info "Reading connection strings ..."
$XML = [xml](Get-Content $ConnectionStringsConfigFilePath)
$Connections = @{}
$XML.ConnectionStrings.add | %{$Connections[$_.name] = $_.connectionString}

#Validates if all SQL connection strings exist in the file
Write-Info "Validating SQL connection strings ..."
$SqlDBConnectionNames | %{
    If($Connections.Keys -notcontains $_)
    {
        Write-Error "'$_' connection string not found." 
    }
    If(!$Connections[$_])
    {
        Write-Error "'$_' connection string is null or empty." 
    }
}

#Validates if the SQL connection strings are valid
$SqlDBConnectionNames | %{
    $ConnectionName = $_
    try
    {
        Write-Info "Checking '$ConnectionName' connection ..."
        Retry-Block {
            Using-Object($Connection = New-Object System.Data.SQLClient.SQLConnection($Connections[$ConnectionName])){
                $Connection.Open() 
                Write-Info "'$ConnectionName' connection is fine."
            }
        }
    }
    catch
    {
        Write-Error "Could not open '$ConnectionName' connection to database" -ErrorAction Continue
        throw
    }
}

#Validate if azure connection string exists in the file
Write-Info "Validating Azure connection string and container ..."
If($Connections.Keys -notcontains $AzureBlobConnectionName)
{
    Write-Error "'$AzureBlobConnectionName' connection string not found." 
}
If(!$Connections[$AzureBlobConnectionName])
{
    Write-Error "'$AzureBlobConnectionName' connection string is null or empty."
}

#Validates if the container exists on the storage account
$AzureConnection= $Connections[$AzureBlobConnectionName]

$Context = New-AzureStorageContext -ConnectionString $AzureConnection

$Container = $null
try
{
    $Container = GetContainer -Context $Context -ContainerName $BlobContainerName
    Write-Info "'$AzureBlobConnectionName' is fine and '$BlobContainerName' container exists."
}
catch
{
    Write-Error "Could not find container '$BlobContainerName'. Make sure the azure storage connection string '$AzureBlobConnectionName' is valid and the container '$BlobContainerName' exists." -ErrorAction Continue
    throw
}
#endregion

#region Status management

$FreshExecution = $False
$Status = $Null
Write-Info "Checking the status ..."
$statusFileName = "status.json"
$ScriptInvocationPath = Split-Path $MyInvocation.MyCommand.Path
$StatusFilePath = (Join-Path $ScriptInvocationPath $statusFileName)

#If the status file doesn't exists, it's a fresh execution
If(!(Test-Path -Path $StatusFilePath -PathType Leaf))
{
    $FreshExecution = $True
}

#If the -Force switch exists, should treat it as a fresh execution
If($Force)
{
    Write-Info "Force switch found. Treating as a fresh execution."

    $FreshExecution = $True
    if(Test-Path $StatusFilePath)
    {
        Remove-Item $StatusFilePath -Force
    }
}

#If it's fresh execution, then the status should be like the following:
If($FreshExecution) 
{
    Write-Info "It's a fresh execution."

    $Status = [Ordered]@{
        "SqlDBConnectionNames" = $SqlDBConnectionNames;
	    "AzureBlobConnectionName"= $AzureBlobConnectionName;
	    "BlobContainerName"= $BlobContainerName;
	    "BlobsCopied"= @();
        "BlobValuesChanged"=@();
    }
}
else
{
    #If it's not a fresh execution, then we should validate the status file:
    Write-Info "It's a continuing execution. Recovering from the last execution ..."
    Write-Info "Validating the status ..."

    #Validate if status file is a valid json file
    $Status = (Get-Content -Raw $StatusFilePath | ConvertFrom-Json -ErrorAction Stop)
    If(!$Status) 
    {
        Write-Error "Status file '$statusFileName' is not valid."
    }
   
    #Validate if status file has AzureBlobConnectionName
    If(!($Status.AzureBlobConnectionName)) 
    {
        Write-Error "'AzureBlobConnectionName' doesn't exists in the status file."
    }

    #Validate if the value which is passed to the script for AzureBlobConnectionName parameter, is same as the value which has been passed in the previous run
    If(!($Status.AzureBlobConnectionName -ceq $AzureBlobConnectionName))
    {
        Write-Error "'AzureBlobConnectionName' for previous execution is '$($Status.AzureBlobConnectionName)'. Pass the same value to the script."
    }

    #Validate if status file has BlobContainerName
    If(!($Status.BlobContainerName))
    {
        Write-Error "'BlobContainerName' doesn't exists in the status file."
    }

    #Validate if the value which is passed to the script for BlobContainerName parameter, is same as the value which has been passed in the previous run
    If(!($Status.BlobContainerName -ceq $BlobContainerName))
    {
        Write-Error "'BlobContainerName' for previous execution was '$($Status.BlobContainerName)'. Pass the same value to the script."
    }

    #Validate if status file has SqlDBConnectionNames
    If(!($Status.SqlDBConnectionNames)) 
    {
        Write-Error "'SqlDBConnectionNames' doesn't exists in the status file."
    }
    
    #Validate if the value which is passed to the script for SqlDBConnectionNames parameter, is same as the value which has been passed in the previous run
    $diff1 = $SqlDBConnectionNames | where {$Status.SqlDBConnectionNames -notcontains $_}
    $diff2 = $Status.SqlDBConnectionNames | where {$SqlDBConnectionNames -notcontains $_}
    If($diff1 -or $diff2)
    {
        Write-Error "'SqlDBConnectionNames' for the previous execution was `($($Status.SqlDBConnectionNames -join ',')`). Pass the same value to the script."
    }

    #Validate if the values in the status file are consistent.
    $diff1 = $Status.BlobsCopied.Name | where {$Status.SqlDBConnectionNames -notcontains $_}
    $diff2 = $Status.BlobValuesChanged.Name | where {$Status.SqlDBConnectionNames -notcontains $_}
    $diff3 = $Status.BlobValuesChanged.Name | where {$Status.BlobsCopied.Name -notcontains $_}
    If($diff1 -or $diff2 -or $diff3)
    {
        Write-Error "Values in 'SqlDBConnectionNames' and 'BlobsCopied' and 'BlobValuesChanged' are inconsistent."
    }
}
#endregion

#region Copy blobs

#Get list of all blob Ids page by page
$PageSize = 500
$TempFolderName = "temp"
$ScriptInvocationPath = Split-Path $MyInvocation.MyCommand.Path
$TempFolderPath = (Join-Path $ScriptInvocationPath $TempFolderName)
If(Test-Path $TempFolderPath)
{
    Remove-Item $TempFolderPath -Recurse -Force
}
New-Item -ItemType Directory -Force -Path $TempFolderPath | Out-Null

Write-Info "`n"

foreach($Connection in $Status.SqlDBConnectionNames)
{
	Write-Info "`n"
	Write-Info "Processing '$Connection' database blobs ..."
    Write-Info "-------------------------------------------"
    $ConnectionString = $Connections[$Connection]
    If($Status.BlobsCopied.Name -contains $Connection)
    {
        Write-Info "'$Connection' blobs has already been processed."
    }
    else
    {
		$BlobIdsCount = GetBlobIdsCount($ConnectionString)
        $PageCount = [Math]::Ceiling([double]$BlobIdsCount/[double]$PageSize)
        $BlobsCopied = 0
        $BlobsProcessing = 0
        for ($PageNo=1; $PageNo -le $PageCount; $PageNo++) 
        {
            $BlobIds = (GetBlobIdsByPage $ConnectionString $PageNo $PageSize).BlobId
            foreach($BlobId in $BlobIds)
            {
                $BlobsProcessing++
                Write-Info "Processing $BlobsProcessing of $BlobIdsCount - '$BlobId'"
                # If blob doesn't exist in azure, download blob from SQL database, then upload it to azure and remove the temp file
                Write-Info "Check if blob exists in azure ..."
                $Blob = BlobExistsInAzure -BlobId $BlobId -ContainerName $BlobContainerName -Context $Context
                if (!$Blob)
                {
                    Write-Info "Blob Doesn't exist in azure."
                    Write-Info "Check if the blob exists in SQL Server ..."
                    $BlobExists = BlobExistsInSql -ConnectionString $ConnectionString -BlobId $BlobId
                    If($BlobExists)
                    {
                        Write-Info "Blob exists on SQL Server."
                        $BlobSize = GetBlobSizeInSql -ConnectionString $ConnectionString -BlobId $BlobId
                        Write-Info "Downloading blob Size = $BlobSize Bytes from SQL..."
                        $File = DownloadBlobFromSql -ConnectionString $ConnectionString -BlobId $BlobId -Folder $TempFolderPath
                        Write-Info "Download blob from SQL completed."
                        $FileSize =(Get-Item $File).Length
                        If($BlobSize -ne $FileSize)
                        {
                            Write-Error "Blob size of '$BlobId' in database is different from the downloaded file size."
                        }
                        Write-Info "Uploading blob having Size = $FileSize to Azure ..."
                        $Blob = UploadBlobToAzure -Context $Context -ContainerName $BlobContainerName -File $File -BlobId $BlobId -Force 
                        Write-Info "Upload blob to Azure completed."
                        $CloudBlobSize = $Blob.ICloudBlob.Properties.Length
                        If($CloudBlobSize -ne $FileSize)
                        {
                            Write-Error "Downloaded blob size is different from the blob size uploaded for '$BlobId'."
                        }
                        $BlobsCopied++

                        Remove-Item $File -Force | Out-Null
                    }
                    else
                    {
                        Write-Info "Warning: '$BlobId' doesn't exist in SQL Server. There is no blob to upload to Azure."
                    }
                }
                else
                {
                    Write-Info "Blob exists in azure. No need to copy from SQL database to Azure."
                }
                Write-Info "`n"
            }
        }

        Write-Info "Processed $BlobIdsCount blobs of '$Connection' database."
        Write-Info "Copied $BlobsCopied blobs. Rest of them were existing on the destination."

        #Update the status file with the new stable status, blobs copied:
        $Status.BlobsCopied += [PSCustomObject]@{
            "Name" = $Connection;
            "Items" = $BlobsCopied;
        }
        $Status | ConvertTo-Json | Out-File -FilePath $StatusFilePath -Encoding unicode -Force
    }
}
foreach($Connection in $Status.SqlDBConnectionNames)
{
	Write-Info "`n"
	Write-Info "Processing '$Connection' database blob fields ..."
    Write-Info "-------------------------------------------------"
    $ConnectionString = $Connections[$Connection]
    If($Status.BlobValuesChanged.Name -contains $Connection)
    {
        Write-Info "'$Connection' blob ids has already been processed."
    }
    else
    {
        Write-Info "Updating '$Connection' database blob ids ..."
        $UpdatedRows = UpdateBlobIds $ConnectionString
        Write-Info "Updated $UpdatedRows records in '$Connection' database."

        #Update the status file with the new stable status, rows updated copied:
        $Status.BlobValuesChanged += [PSCustomObject]@{
            "Name" = $Connection;
            "Items" = $UpdatedRows;
        }
        $Status | ConvertTo-Json | Out-File -FilePath $StatusFilePath -Encoding unicode -Force
    }
}
#endregion

#region Write summary and Cleanup
Write-Info "`n"

Write-Info "----------------------------------------"
Write-Info "Invoke-Migration completed successfully."
Write-Info "$(Get-Date)"
Write-Info "----------------------------------------"

Write-Info "Processed databases: `($($Status.SqlDBConnectionNames -join ',')`)"
Write-Info "Blobs:"
foreach($element in $Status.BlobsCopied)
{
    Write-Info "    $($element.Name): $($element.Items) blobs copied to azure."
}
Write-Info "Blob fields:"
foreach($element in $Status.BlobValuesChanged)
{
    Write-Info "    $($element.Name): $($element.Items) blob fields updated."
}

Write-Info "----------------------------------------"

If(Test-Path $StatusFilePath)
{
    Remove-Item $StatusFilePath -Force
}
If(Test-Path $TempFolderPath)
{
    Remove-Item $TempFolderPath -Recurse -Force
}
#endregion
}
Catch [System.IO.IOException]
{
    Write-Error "Another instance of the script is running." -ErrorAction Continue
    throw
}
Finally
{
    if($Lock)
    {
        $Lock.Dispose();
    }
    If(Test-Path $LockFilePath)
    {
        Remove-Item $LockFilePath -Force
    }
}